# -*- coding: utf-8 -*-
#
# This software may be modified and distributed under the terms
# of the MIT license.  See the LICENSE file for details.

__version__ = '2.0.0'

# When using an in-memory only cache, this persists the cache through
# thread failures, shutdowns, and restarts.
EVENT_CACHE = {}
